#!/usr/bin/perl
######################################
# Align seq against PDBs by TMalign  #
######################################

$librarydir=$ARGV[0];
$init=$ARGV[1];
$strname=$ARGV[2];
$ndecoy=$ARGV[3];
$rmsdcut=$ARGV[4];
$hbcut=0.80;
$bins=10;

%ts=(
     'GLY'=>'G',
     'ALA'=>'A',
     'VAL'=>'V',
     'LEU'=>'L',
     'ILE'=>'I',
     'SER'=>'S',
     'THR'=>'T',
     'CYS'=>'C',
     'MET'=>'M',
     'PRO'=>'P',
     'ASP'=>'D',
     'ASN'=>'N',
     'GLU'=>'E',
     'GLN'=>'Q',
     'LYS'=>'K',
     'ARG'=>'R',
     'HIS'=>'H',
     'PHE'=>'F',
     'TYR'=>'Y',
     'TRP'=>'W',

     'ASX'=>'B',
     'GLX'=>'Z',
     'UNK'=>'X',

     'G'=>'GLY',
     'A'=>'ALA',
     'V'=>'VAL',
     'L'=>'LEU',
     'I'=>'ILE',
     'S'=>'SER',
     'T'=>'THR',
     'C'=>'CYS',
     'M'=>'MET',
     'P'=>'PRO',
     'D'=>'ASP',
     'N'=>'ASN',
     'E'=>'GLU',
     'Q'=>'GLN',
     'K'=>'LYS',
     'R'=>'ARG',
     'H'=>'HIS',
     'F'=>'PHE',
     'Y'=>'TYR',
     'W'=>'TRP',

     'a'=>'CYS',
     'b'=>'CYS',
     'c'=>'CYS',
     'd'=>'CYS',
     'e'=>'CYS',
     'f'=>'CYS',
     'g'=>'CYS',
     'h'=>'CYS',
     'i'=>'CYS',
     'j'=>'CYS',
     'k'=>'CYS',
     'l'=>'CYS',
     'm'=>'CYS',
     'n'=>'CYS',
     'o'=>'CYS',
     'p'=>'CYS',
     'q'=>'CYS',
     'r'=>'CYS',
     's'=>'CYS',
     't'=>'CYS',
     'u'=>'CYS',
     'v'=>'CYS',
     'w'=>'CYS',
     'x'=>'CYS',
     'y'=>'CYS',
     'z'=>'CYS',

     'B'=>'ASX',
     'Z'=>'GLX',
     'X'=>'CYS',
    );

##### decide max number of templates going to be used ######
if($ndecoy>1000){
  $mntuse=int($ndecoy/10);
  $mntuse=200 if $mntuse>200;
}elsif($ndecoy<120){
  $mntuse=$ndecoy;
}else{
  $mntuse=120;
}

########## read structure template list #############
$n=0;
@pdbset=();

$Plst=0;
if(-s "PDBlist"){
  if(-s "PDBlist" > 10){
    open(NAM,"PDBlist") or die "Cannot open PDBlist $!\n";
    while($line=<NAM>){
      $pdbset[$n++]=$1 if $line=~/(\S+)/;
    }
    close(NAM);
    $Plst=1;
  }
}
if($Plst==0){
  opendir(LD, "$librarydir") or die "Can't open:$librarydir $!\n" ;
  @libfiles= readdir(LD);
  closedir(LD);

  foreach $lfl (@libfiles){
    if($lfl=~/(\S+).pdb/){
      $pdbset[$n++]=$1;
    }
  }
}

########## read protein length #############
open(RMSI,"rmsinp") or die "Cannot open rmsinp $!\n";
<RMSI>=~/\d+\s+(\d+)/;
$Lch=$1;
close(RMSI);

####### calculate number of Hbond in input structure #########
$nnHB=0;
%ssdis0=();
if(-s "$strname"){
  @rst=`./stride $strname -h`;
  foreach(@rst){
    if($_=~/^ACC\s+\S+\s+\S+\s+(\d+) ->\s+\S+\s+\S+\s+(\d+)/){
      if(!exists $ssdis0{$1}{$2} && !exists $ssdis0{$2}{$1}){
        $ssdis0{$1}{$2}=$nnHB++;
      }
    }
    if($_=~/^DNR\s+\S+\s+\S+\s+(\d+)\s+\d+ ->\s+\S+\s+\S+\s+(\d+)/){
      if(!exists $ssdis0{$1}{$2} && !exists $ssdis0{$2}{$1}){
        $ssdis0{$1}{$2}=$nnHB++;
      }
    }
  }
}else{
  print "No $strname file found.\n";
  exit();
}

##### run TMalign between input structure and all templates ######
$n_tmp=0;
$ncheck1=0;
$ncheck2=0;
$ncheck3=0;
%rmsds=();
%tmsco=();
%seqid=();
%cover=();
%segmt=();
%seqt=();
%seqq=();
%hbond=();
%lgth=();
foreach $pdb (@pdbset){
  next if length $pdb < 2;
  if(-s "$librarydir/$pdb.pdb"){
    #print "$n_tmp:$pdb\n";
    @rst=`./TMalign $librarydir/$pdb.pdb CA`;
    $flag=0;
    foreach $line (@rst){
      if($line=~/":" denotes aligned residue pairs of/){
        $flag=1;
        next;
      }
      $flag++ if $flag>0;
      if($flag==2){
        $seqt{$pdb}=$line;
        $seqt{$pdb}=~s/\*//mg;
        $seqt{$pdb}=~s/\s+//g;
        $ncheck1++;
      }
      if($flag==3){
        chomp($line);
        $ncp=length $line;
        $segmt{$pdb}=0;
        $segmt{$pdb}++ if substr($line,0,1) ne ' ';
        for($i=1;$i<$ncp;$i++){
          $segmt{$pdb}++ if substr($line,$i-1,1) eq ' ' && substr($line,$i,1) ne ' ';
        }
      }
      if($flag==4){
        $seqq{$pdb}=$line;
        $seqq{$pdb}=~s/\*//mg;
        $seqq{$pdb}=~s/\s+//g;
        $lgth{$pdb}=length $seqq{$pdb};
        $n_tmp++;
        last;
      }
      if($line=~/Aligned length=\s+(\d+), RMSD=\s+(\S+)\s+Seq_ID=n_identical\/n_aligned=\s+(\S+)/){
        $cover{$pdb}=$1;
        $rmsds{$pdb}=$2;
        $seqid{$pdb}=$3;
        $rmsds{$pdb}=~s/,//mg;
        $ncheck2++;
      }
      if($line=~/TM-score=\s+(\S+)\s+\(if normalized by length of Chain_2/){
        $tmsco{$pdb}=$1;
        $ncheck3++;
      }
    }
    if($flag<4){
      print "TMalign output of $pdb VS CA is wrong.\n";
      exit();
    }
    $hbond{$pdb}=nhbond($pdb);
  }else{
    print "No PDB file of $pdb found.\n";
    next;
  }
}
if($ncheck1!=$ncheck2||$ncheck1!=$n_tmp||$ncheck3!=$n_tmp){
  print "Inconsistency: n_tmp=$n_tmp  ncheck1=$ncheck1  ncheck2=$ncheck2  ncheck3=$ncheck3\n";
  exit();
}

########## rank templates by RMSD #############
@rank= keys %rmsds;
$nrank=@rank;
if($nrank!=$n_tmp){
  print "Inconsistency: nrank=$nrank  n_tmp=$n_tmp\n";
  $n_tmp=$nrank;
}
for($i=0;$i<$n_tmp;$i++){
  for($j=$n_tmp-1;$j>$i;$j--){
    if($rmsds{$rank[$j]}<$rmsds{$rank[$j-1]}){
      $tmp=$rank[$j];
      $rank[$j]=$rank[$j-1];
      $rank[$j-1]=$tmp;
    }
  }
}

########## select templates in each bin #############
$rmsdmax=$rmsds{$rank[$n_tmp-1]};
$binwide=$rmsdmax/($bins+0.000001);

###-> decide max number of templates going to be used in each bin.
@ntuse=();
$tnb=int($rmsdcut/$binwide);
if($tnb>$bins){
  $ndext=int($mntuse*$tnb**2/($tnb**2-($tnb-$bins)**2));
}else{
  $bins=$tnb;
  $ndext=$mntuse;
}
if($ndext>$tnb){
  $mnd=int($ndext*2.0/$tnb)-1;
  $nad=($mnd-1.0)/($tnb-1.0);
  for($i=0;$i<$bins;$i++){
    $ntuse[$i]=$mnd-int($nad*$i);
  }
}else{
  for($i=0;$i<$bins;$i++){
    $ntuse[$i]=1;
  }
}

###-> decide coverage cutoff.
if($rmsdmax<$rmsdcut){
  $cvcut1=($rmsdmax*$Lch/($rmsdcut*200.0))**0.333;
}else{
  $cvcut1=0.8*($Lch/200.0)**0.333;
}
$cvcut1=0.50 if $cvcut1<0.50;
$cvcut1=0.80 if $cvcut1>0.80;

###-> calculate number of templates within RMSD cutoff.
$n_cut=0;
while($rmsdcut>=$rmsds{$rank[$n_cut]}){
  $n_cut++;
  last if $n_cut>$n_tmp-1;
}

###-> decide the rate of template selection in following two steps.
$trate=sqrt($rmsdmax/$rmsdcut);
$trate=1.0 if $trate>1.0 || $rmsdcut<6.0;
$trate=0.5 if $trate<0.5;

###->step 1: select good templates for generating relatively low RMSD decoys.
$bin=1;
$n=0;
$segmx=0;
$tot1=0;
$totnb1=0;
@cvrk=();
%slcts=();
$nHBcut=$nnHB*$hbcut;
@ntu1=();
@n_eachbin1=();
for($i=0;$i<$bins;$i++){
  $ntu1[$i]=int($ntuse[$i]*$trate);
  $ntu1[$i]++ if $ntuse[$i]*$trate>$ntu1[$i];
  $tot1+=$ntu1[$i];
  $n_eachbin1[$i]=0;
}
for($i=0;$i<$n_cut;$i++){
  next if $rmsds{$rank[$i]}<0.001;
  if($rmsds{$rank[$i]}<$binwide*$bin){
    if($cover{$rank[$i]}/($Lch+0.00001)>=$cvcut1 && $hbond{$rank[$i]}>=$nHBcut){
      $cvrk[$n++]=$rank[$i];
      $segmx=$segmt{$rank[$i]} if $segmt{$rank[$i]}>$segmx;
    }
  }else{
    if($n>0){
      &coeval($n) if $n>1;
      if($ntu1[$bin-1]>=$n){
        for($ii=0;$ii<$n;$ii++){
          $slcts{$cvrk[$ii]}=$rmsds{$cvrk[$ii]};
        }
        $n_eachbin1[$bin-1]=$n;
        $totnb1+=$n;
      }else{
        for($ii=0;$ii<$ntu1[$bin-1];$ii++){
          $slcts{$cvrk[$ii]}=$rmsds{$cvrk[$ii]};
        }
        $n_eachbin1[$bin-1]=$ntu1[$bin-1];
        $totnb1+=$ntu1[$bin-1];
      }
      $n=0;
      $segmx=0;
      @cvrk=();
    }
    $bin++;
    $i--;
    last if $bin>$bins;
  }
}

###->step 2: select templates for generating high RMSD decoys.
$cvcut2=$cvcut1;
while($trate<1.0){
  $tot2=0;
  $totnb2=0;
  @ntu2=();
  @n_eachbin2=();
  %slc=();
  for($i=0;$i<$bins;$i++){
    $ntu2[$i]=int($ntuse[$bins-$i-1]*(1.0-$trate));
    $tot2+=$ntu2[$i];
    $n_eachbin2[$i]=0;
    $nt2[$i]=0;
    $n2[$i]=1;
  }
  $ntu2[$bins-1]=1 if $ntu2[$bins-1]<=0;

  $coverup=1.0;
  $ncvbin=$ntu2[$bins-1];
  $coverbw=($coverup-$cvcut2)/$ncvbin;
  $coverdn=$coverup-$coverbw;
  while($coverdn>=$cvcut2){
    $bin=1;
    $n=0;
    $segmx=0;
    for($i=0;$i<$n_cut;$i++){
      next if $rmsds{$rank[$i]}<0.001;
      if($rmsds{$rank[$i]}<$binwide*$bin){
        $cvp=$cover{$rank[$i]}/($Lch+0.00001);
        if($cvp>$coverdn && $cvp<=$coverup){
          $cvrk[$n++]=$rank[$i];
          $segmx=$segmt{$rank[$i]} if $segmt{$rank[$i]}>$segmx;
        }
      }else{
        $nt2[$bin-1]++;
        if($ntu2[$bin-1]>0){
          $nts=int($ncvbin*$n2[$bin-1]/$ntu2[$bin-1]);
          if($nts==$nt2[$bin-1]){
            $n2[$bin-1]++;
            if($n>0){
              &coeval($n) if $n>1;
              if(!exists $slcts{$cvrk[0]}){
                $slc{$cvrk[0]}=$rmsds{$cvrk[0]};
                $n_eachbin2[$bin-1]++;
                $totnb2++;
              }
            }
          }
        }
        $n=0;
        $segmx=0;
        @cvrk=();
        $bin++;
        $i--;
        last if $bin>$bins;
      }
    }
    $coverup=$coverdn;
    $coverdn=$coverdn-$coverbw;
  }
  if($totnb2/($tot2+0.00001)<0.27 && $cvcut2>0.02){
    $cvcut2-=0.02;
  }else{
    last;
  }
}
@slcs=keys %slc;
foreach(@slcs){
  $slcts{$_}=$slc{$_};
}

@slct=sort {$slcts{$a} <=> $slcts{$b}} keys %slcts;
$ntmp=@slct;

########## generate init.dat #############
open(INI,"> $init") or die "Can't make $init $!\n";
if($ntmp>0){
  $n_tot=$ntmp;
  $n_tot=$n_tot+1 if $rmsds{$slct[0]}>0.001;
}else{
  $n_tot=1;
}
printf INI "%8d  %6.2f  SEQ\n",$n_tot,$rmsdmax;

open(CA,"CA") or die "No input structure CA\n";
print INI " $Lch  0.00  1  native  CA\n";
while($line=<CA>){
  print INI $line if $line=~/^ATOM/;
}
print INI "TER  \n";
close(CA);

%seqTMP=();
%numT=();
%x=();
%y=();
%z=();
$nn=2;
for($i=0;$i<$ntmp;$i++){
  open(TMP,"$librarydir/$slct[$i].pdb");
  $n=0;
  while($line=<TMP>){
    if($line=~/^ATOM/){
      $atom=substr($line,12,4);
      $atom=~s/\s//mg;
      if($atom eq "CA"){
        $n++;
        $seqTMP{$n}=$ts{substr($line,17,3)}; #only for check
        $numT{$n}=substr($line,22,4);
        $x{$n}=substr($line,30,8);
        $y{$n}=substr($line,38,8);
        $z{$n}=substr($line,46,8);
      }
    }
  }
  close(TMP);

  printf INI "%8d%8.2f  %d  %s  TMalign\n",$cover{$slct[$i]},$rmsds{$slct[$i]},$nn++,$slct[$i];
  $iQ=0;
  $iT=0;
  for($j=1;$j<=$lgth{$slct[$i]};$j++){
    $sQ=substr($seqq{$slct[$i]},$j-1,1);
    $sT=substr($seqt{$slct[$i]},$j-1,1);
    if($sQ eq "-"){
      $iT++;
    }
    if($sT eq "-"){
      $iQ++;
    }
    if($sQ ne "-" && $sT ne "-"){
      $iQ++;
      $iT++;
      printf INI "ATOM  %5s  CA  %3s  %4d    %8.3f%8.3f%8.3f%5s %3s\n",
      $iQ,$ts{$sQ},$iQ,$x{$iT},$y{$iT},$z{$iT},$numT{$iT},$ts{$sT};
      if($sT ne $seqTMP{$iT}){
        print "$slct[$i]: $iT - $sT = $seqTMP{$iT}\n";
      }
    }
  }
  printf INI "TER\n";
}
close(INI);

########## output for check #############
open(OUT,"> initcheck.dat") or die "Can't make initcheck.dat $!\n";
printf OUT "TMP chosen for init.dat :\n";
printf OUT "Lch=%-6drmsdmax=%-7.2fnnHB=%-8dtrate=%-4.2f\n",$Lch,$rmsdmax,$nnHB,$trate;
printf OUT "ntmp=%-5dmntuse=%-8dcvcut1=%-6.2fcvcut2=%-4.2f\n",$ntmp,$mntuse,$cvcut1,$cvcut2;
printf OUT "TMP_NAM     RMSD  TMscore   COVER    NSEG    N_HB\n";
$armsds=$atmsco=$acover=$asegmt=$ahbond=0.0;
for($i=0;$i<$ntmp;$i++){
  printf OUT "%-8s%8.2f%8.3f%8d%8d%8d\n",$slct[$i],$rmsds{$slct[$i]},$tmsco{$slct[$i]},$cover{$slct[$i]},$segmt{$slct[$i]},$hbond{$slct[$i]};
  $armsds+=$rmsds{$slct[$i]};
  $atmsco+=$tmsco{$slct[$i]};
  $acover+=$cover{$slct[$i]};
  $asegmt+=$segmt{$slct[$i]};
  $ahbond+=$hbond{$slct[$i]};
}
$armsds/=$ntmp;
$atmsco/=$ntmp;
$acover/=$ntmp;
$asegmt/=$ntmp;
$ahbond/=$ntmp;
printf OUT "Average %8.2f%8.3f%8d%8d%8d\n",$armsds,$atmsco,$acover,$asegmt,$ahbond;
printf OUT "**************************************************\n";
printf OUT "EACH-BIN       No.of.TMP(step1)  No.of.TMP(step2)\n";
for($i=0;$i<$bins;$i++){
  printf OUT "[%5.2f-%-5.2f]%9d/%-9d%9d/%-9d\n",$i*$binwide,($i+1)*$binwide,$n_eachbin1[$i],$ntu1[$i],$n_eachbin2[$i],$ntu2[$i];
}
printf OUT "TOTAL        %9d/%-9d%9d/%-9d\n",$totnb1,$tot1,$totnb2,$tot2;
printf OUT "**************************************************\n";
printf OUT "No. of TMP within rmsdcut = %-8d\n",$n_cut;
printf OUT "TMP_NAM     RMSD  TMscore   COVER    NSEG    N_HB\n";
for($i=0;$i<$n_cut;$i++){
  printf OUT "%8s%8.2f%8.3f%8d%8d%8d\n",$rank[$i],$rmsds{$rank[$i]},$tmsco{$rank[$i]},$cover{$rank[$i]},$segmt{$rank[$i]},$hbond{$rank[$i]};
}
close(OUT);

########## output for webserver #############
open(TMP,"> tmpl.dat") or die "Can't make tmpl.dat $!\n";
printf TMP ">>------------------ Structure Templates used by 3DRobot ------------------<<\n\n";
for($i=0;$i<$ntmp;$i++){
  printf TMP ">No.%-4dPDBID: $slct[$i]\n",$i+1;
  printf TMP "Coverage=%-5dSeq_ID=%-7.2fTM-score=%-7.3fRMSD=%-5.2f\n",$cover{$slct[$i]},$seqid{$slct[$i]},$tmsco{$slct[$i]},$rmsds{$slct[$i]};
  printf TMP "T:$seqt{$slct[$i]}\n";
  printf TMP "Q:$seqq{$slct[$i]}\n\n";
}
close(TMP);


sub nhbond{
  $p="$_[0]";
  %nresn=();
  %ssdis0=();
  $NHB0=0;
  $NHB=0;
  @results=`./stride $librarydir/$p.pdb -h`;
  if(@results<=1){#avoid stride bug
    $result=`cat $librarydir/$p.pdb`;
    $result=~s/^HEADER//mg;
    open(TMPL,"> $p.pdb");
    print TMPL $result;
    close(TMPL);
    @results=`./stride $p.pdb -h`;
    `rm $p.pdb`;
  }

  foreach $result (@results){
    if($result=~/^ACC\s+\S+\s+\S+\s+\d+\s+(\d+) ->\s+\S+\s+\S+\s+\d+\s+(\d+)/){
      $i=$1+1;
      $j=$2+1;
      if(!exists $ssdis0{$i}{$j} && !exists $ssdis0{$j}{$i}){
        $ssdis0{$i}{$j}=$NHB0++;
      }
    }
    if($result=~/^DNR\s+\S+\s+\S+\s+\d+\s+(\d+) ->\s+\S+\s+\S+\s+\d+\s+(\d+)/){
      $i=$1+1;
      $j=$2+1;
      if(!exists $ssdis0{$i}{$j} && !exists $ssdis0{$j}{$i}){
        $ssdis0{$i}{$j}=$NHB0++;
      }
    }
  }

  $iQ=0;
  $iT=0;
  for($j=1;$j<=$lgth{$p};$j++){
    $sQ=substr($seqq{$p},$j-1,1);
    $sT=substr($seqt{$p},$j-1,1);
    if($sQ eq "-"){
      $iT++;
    }
    if($sT eq "-"){
      $iQ++;
    }
    if($sQ ne "-" && $sT ne "-"){
      $iQ++;
      $iT++;
      $nresn{$iT}=$iQ;
    }
  }

  @pair_i= keys %ssdis0;
  foreach $i (@pair_i){
    @pair_j= keys %{$ssdis0{$i}};
    foreach $j (@pair_j){
      if(exists $nresn{$i} && exists $nresn{$j}){
        $NHB++;
      }
    }
  }
  return $NHB;
}

sub coeval{
  $n=$_[0];
  @coeva=();
  for($ii=0;$ii<$n;$ii++){
    $coeva[$ii]=($hbond{$cvrk[$ii]}/($nnHB+0.00001))*0.5
               +($cover{$cvrk[$ii]}/($Lch+0.00001))*0.25
               +(($segmx-$segmt{$cvrk[$ii]})/($segmx+0.00001))*0.25
  }
  for($ii=0;$ii<$n;$ii++){
    for($jj=$n-1;$jj>$ii;$jj--){
      if($coeva[$jj]>$coeva[$jj-1]){
        $coe=$coeva[$jj];
        $coeva[$jj]=$coeva[$jj-1];
        $coeva[$jj-1]=$coe;

        $tmp=$cvrk[$jj];
        $cvrk[$jj]=$cvrk[$jj-1];
        $cvrk[$jj-1]=$tmp;
      }
    }
  }
}