#!/usr/bin/perl -w
########################
# Decoy set evaluation #
########################

$strname=$ARGV[0];
$rmsdcut=$ARGV[1];
$binwd=$ARGV[2];
$ss=$ARGV[3];
$outdir=$ARGV[4];
$lst="list.txt";
$outnam="rst.dat";
$intvl=5;
$nbin=int($rmsdcut/$binwd);

%ts=(
     'GLY'=>'G',
     'ALA'=>'A',
     'VAL'=>'V',
     'LEU'=>'L',
     'ILE'=>'I',
     'SER'=>'S',
     'THR'=>'T',
     'CYS'=>'C',
     'MET'=>'M',
     'PRO'=>'P',
     'ASP'=>'D',
     'ASN'=>'N',
     'GLU'=>'E',
     'GLN'=>'Q',
     'LYS'=>'K',
     'ARG'=>'R',
     'HIS'=>'H',
     'PHE'=>'F',
     'TYR'=>'Y',
     'TRP'=>'W',

     'ASX'=>'B',
     'GLX'=>'Z',
     'UNK'=>'X',
    );

%sstran=(
     1=>'C',
     2=>'H',
     4=>'E',
    );

@evnss=();
@criteria=qw(secps rgyrt cntnm);
$ncrit=@criteria;
@dname=@rmsds=@secps=@rgyrt=@cntnm=();

open(OUT,"> $outnam") or die "Can't make:$outnam $!\n";

open(RMSI,"rmsinp") or die "Cannot open rmsinp $!\n";
<RMSI>=~/\d+\s+(\d+)/;
$Lch=$1;
close(RMSI);

######## calculate native second structure infomation #########
$na=0;
$nb=0;
$ntt=0;
$sequence='';
$seconstr='';
@results=`./stride $outdir/$strname`;
if(@results<$Lch){
  print "Output error by running stride on $strname. Quit!\n";
  exit();
}
foreach $result (@results){
  if($result=~/ASG\s+(\S+)\s+\S+\s+\d+\s+\d+\s+(\S+)/){
    $sequence.=$ts{$1};
    $j=$2;
    $ntt++;
    if($j eq "H"||$j eq "G" ||$j eq "I"){
      $sc=2;
      $na++;
    }
    elsif($j eq "B" ||$j eq "b" ||$j eq "E"){
      $sc=4;
      $nb++;
    }else{
      $sc=1;
    }
    $seconstr.=$sstran{$sc};
  }
}
if($ntt!=$Lch){
  printf "Protein length inconsistence in rmsinp and $strname.\n";
  exit;
}

########### decide class ------------>
$class='';
if($na<=10 && $nb>=11){
  $class='Beta protein';
}elsif($na>=11 && $nb<=6){
  $class='Alpha protein';
}elsif($na>=11 && $nb>=7){
  $class='Alpha and Beta protein';
}

$spct=($na+$nb)*100.0/$Lch;
&strinfo("$outdir/$strname");
$nvalue[0]=$spct;
$nvalue[1]=$rgyr;
$nvalue[2]=$ncnt;

######## output native structure infomation ###########
printf OUT ">>-------------------- Decoy set detailed informations --------------------<<\n";
if(length $class >1){
printf OUT "ID:%12s\nLength:%5s      Class: $class\n\n",$ss,$Lch;
}else{
  printf OUT "ID:%12s       Length:%5s\n\n",$ss,$Lch;
}
printf OUT "Protein sequence and second structure: (H->helix;E->sheet;C->coil)\n";
printf OUT "$sequence\n";
printf OUT "$seconstr\n\n";

################## read decoy set list #####################
open(LIST,"$outdir/$lst") or die "Can't open:$outdir/$lst $!\n";
@list=<LIST>;
close(LIST);

$totrms=0;
$ndecoy=$ndcut=0;
for($i=0;$i<=$nbin;$i++){
  $evnss[$i]=0;
}
foreach(@list){
  if($_=~/(\S+)\s+(\S+)/){
    $decoy=$1;
    $rmsd=$2;
    next if $decoy=~/NAME/;
    if(-s "$outdir/$decoy" > 100){

      $spct=&secpct("$outdir/$decoy");
      &strinfo("$outdir/$decoy");

      if($rmsd>0.0001){
        $ndecoy++;
        push(@dname,$decoy);
        push(@rmsds,$rmsd);
        push(@secps,$spct);
        push(@rgyrt,$rgyr);
        push(@cntnm,$ncnt);
        if($rmsd<$rmsdcut){
          $totrms+=$rmsd;
          $ndcut++;
          $evnss[int($rmsd/$binwd)]++;
        }else{
          $evnss[$nbin]++;
        }
      }
    }else{
      print "Warning: $decoy is missing. Omit!\n";
    }
  }
}

################## calculate evenness #####################
$evns=0.0;
$aver=$ndcut/($nbin+0.000001);
for($i=0;$i<$nbin;$i++){
  $evns+=($evnss[$i]-$aver)**2;
}
$evns=sqrt($evns)/sqrt(($aver*($nbin-1.0))**2+($nbin-1.0)*$aver**2);

################## calculate pair-wise RMSD #####################
$result=`./pwRMSD $outdir $lst $rmsdcut`;
if($result=~/Nor_pair_wise_RMSD=\s+(\S+)/){
  $npwRMSD=$1;
}else{
  print "Error when calculating pwRMSD. Quit!\n";
  exit();
}

################## output decoy informations #####################
printf OUT "EVENNESS:%8.3f  npwRMSD:%8.3f\n",1-$evns,$npwRMSD;

printf OUT "RMSD-BIN";
for($i=0;$i<$nbin;$i++){
  if($i<10){
    printf OUT "  >$i<";
  }else{
    printf OUT " >$i<";
  }
}
printf OUT "  >$nbin TOTAL\nN_Decoy ";
for($i=0;$i<=$nbin;$i++){
  printf OUT "%5d",$evnss[$i];
}
printf OUT "%6d\n\n",$ndecoy;

printf OUT "Name                  RMSD SEC_PCT R_GYRAT CNT_NUM\n";
printf OUT "%-18s    0.00%8.2f%8.2f%8.3f\n",$strname,$nvalue[0],$nvalue[1],$nvalue[2];
for($i=0;$i<$ndecoy;$i++){
  printf OUT "%-18s%8.2f%8.2f%8.2f%8.3f\n",
    $dname[$i],$rmsds[$i],$secps[$i],$rgyrt[$i],$cntnm[$i];
}

if(@rmsds>20){
  $averms=$totrms/$ndecoy;

  $rsigma=0.0;
  for($i=0;$i<$ndecoy;$i++){
    $rsigma+=($rmsds[$i]-$averms)**2; 
  }

  for($i=0;$i<$ncrit;$i++){
    $crit=$criteria[$i];

    $nbigger=0;
    $nsmaller=0;
    $totval=0.0;
    for($j=0;$j<$ndecoy;$j++){
      $nbigger++ if $$crit[$j]>=$nvalue[$i];
      $nsmaller++ if $$crit[$j]<=$nvalue[$i];
      $totval+=$$crit[$j];
    }
    $aveval[$i]=$totval/$ndecoy;

    if($nbigger<1 or $nsmaller<1){
      $top1[$i]='Yes';
    }else{
      $top1[$i]='No';
    }
    if($nbigger<5 or $nsmaller<5){
      $top5[$i]='Yes';
    }else{
      $top5[$i]='No';
    }

    $sigma=0.0;
    $prson[$i]=0.0;
    for($j=0;$j<$ndecoy;$j++){
      $sigma+=($$crit[$j]-$aveval[$i])**2; 
      $prson[$i]+=($$crit[$j]-$aveval[$i])*($rmsds[$j]-$averms);
    }
    $zsco[$i]=($nvalue[$i]-$aveval[$i])/sqrt($sigma/$ndecoy);
    $prson[$i]/=sqrt($sigma*$rsigma);
  }
  printf OUT "Average           %8.3f%8.2f%8.2f%8.3f\n",$averms,$aveval[0],$aveval[1],$aveval[2];
  printf OUT "Pcorr-to-dcyRMSD       -  %8.2f%8.2f%8.3f\n",$prson[0],$prson[1],$prson[2];
  printf OUT "Native-rank-in-top1    -  %8s%8s%8s\n",$top1[0],$top1[1],$top1[2];
  printf OUT "Native-rank-in-top5    -  %8s%8s%8s\n",$top5[0],$top5[1],$top5[2];
  printf OUT "Native-Zscore          -  %8.3f%8.3f%8.3f",$zsco[0],$zsco[1],$zsco[2];
}

close(OUT);
`rm $outnam` if -s $outnam <100;

`cp $outnam $outdir/` if -s $outnam;


################## program end #####################

sub strinfo{
  $i=0;
  @x=@y=@z=();
  open(DPDB,"$_[0]") or die "Can't open $_[0] $!\n";
  while(<DPDB>){
    if($_=~/^ATOM/){
      if(substr($_,13,2) eq 'CA'||substr($_,12,2) eq 'CA'||substr($_,14,2) eq 'CA'){
        $x[$i] = substr($_, 30, 8);
        $y[$i] = substr($_, 38, 8);
        $z[$i++] = substr($_, 46, 8);
      }
    }elsif ($_=~/^TER/ && $i>0){
      last;
    }
  }
  $Lch=$i;
  close(DPDB);

  $m=0;
  $ncnt=0;
  $rgyr=0.0;
  $cx=$cy=$cz=0.0;
  for($i=0;$i<$Lch;$i++){
    $cx+=$x[$i];
    $cy+=$y[$i];
    $cz+=$z[$i];
  }
  $cx/=$Lch;
  $cy/=$Lch;
  $cz/=$Lch;

  for($i=0;$i<$Lch;$i++){
    $rgyr+=($x[$i]-$cx)**2+($y[$i]-$cy)**2+($z[$i]-$cz)**2;
  }
  $rgyr=sqrt($rgyr/$Lch);

  for($i=0;$i<$Lch-$intvl-1;$i++){
    for($j=$i+$intvl+1;$j<$Lch;$j++){
      $r=sqrt(($x[$i]-$x[$j])**2+($y[$i]-$y[$j])**2+($z[$i]-$z[$j])**2);
      if($r<=8){
        $ncnt++;
      }
      $m++;
    }
  }
  $ncnt=$ncnt*100.0/$m;
}

sub secpct{
  $nsec=0;
  $totsec=0;
  @results=`./stride $_[0]`;
  foreach $result (@results){
    if($result=~/ASG\s+\S+\s+\S+\s+\d+\s+\d+\s+(\S+)/){
      $j=$1;
      if($j eq "H"||$j eq "G" ||$j eq "I"){
        $sc=2;
      }
      elsif($j eq "B" ||$j eq "b" ||$j eq "E"){
        $sc=4;
      }else{
        $sc=1;
      }
      if($sc>1){
        $nsec++;
      }
      $totsec++;
    }
  }
  $spct=$nsec*100.0/$totsec;
}