#!/usr/bin/perl -w

#########################################################################
$user      = "your_user_name";             #useful when making tmp dir
$pkgdir    = "/home/$user/3DRobot1.0";     #directory containing README.txt
$libdir    = "$pkgdir/PDBlib";             #structure template library directory
$strname   = "strname";                    #input PDB structure name
$indir     = "$pkgdir/data";               #where the input structure places.
$outdir    = "$pkgdir/data";               #where the final decoys are saved
$runstyle  = "serial";                     #parallel or serial
$npara     = 10;                           #Number of parallel jobs allowed in simulation/refinement.
$ndecoy    = 100;                          #Number of output decoys (Recommended range:[10,1000]).
$rmsdcut   = 12;                           #RMSD cutoff for output decoys (Recommended ragne [5,20]).
$hours     = 48;                           #Maximum runing time
$check     = "false";                      #true, check and only run unfinished steps; false, run all steps
$rcd       = "false";                      #whether keep recod&intermediate files.
$argshow   = "true";                       #true, show the setting informations, false, do not show them


$system    = $^O;
unless (lc($system) eq "linux"){
    printf("Your operating system $system is unsupported at this time\n");
    printf("Currently only Linux is supported\n");
    exit();
}

#################### get arguments ######################
$totarg=0;
foreach (@ARGV){
    $totarg++;
}
if($totarg<3){
  &print_help();
  exit();
}
$flagmand[0]=0;
$flagmand[1]=0;
$flagmand[2]=0;
$flagmand[3]=0;
$flagmand[4]=0;


for($i=0;$i<$totarg-1;$i++){
  $str1=substr($ARGV[$i+1],0,1);
  if($str1 eq "-"){
    next;
  }
  if($ARGV[$i] eq "-pkgdir"){
    $pkgdir=$ARGV[$i+1];
    $pkgdir=~s/\/$//;
    $flagmand[0]=1;
  }
  elsif($ARGV[$i] eq "-indir"){
    $indir=$ARGV[$i+1];
    $indir=~s/\/$//;
    $flagmand[1]=1;
  }
  elsif($ARGV[$i] eq "-pdb"){
    $strname=$ARGV[$i+1];
    $flagmand[2]=1;
  }
  elsif($ARGV[$i] eq "-libdir"){
    $libdir=$ARGV[$i+1];
    $libdir=~s/\/$//;
    $flagmand[3]=1;
  }
  elsif($ARGV[$i] eq "-outdir"){
    $outdir=$ARGV[$i+1];
    $outdir=~s/\/$//;
    $flagmand[4]=1;
  }
  elsif($ARGV[$i] eq "-run"){
    $runstyle=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-npara"){
    $npara=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-nd"){
    $ndecoy=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-cut"){
    $rmsdcut=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-check"){
    $check=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-hour"){
    $hours=$ARGV[$i+1];
    $hours=6 if $hours<6;
  }
  elsif($ARGV[$i] eq "-record"){
    $rcd=$ARGV[$i+1];
  }
  elsif($ARGV[$i] eq "-args"){
    $argshow=$ARGV[$i+1];
  }
}

if($flagmand[0]==0){
  &print_help();
  print "please set -pkgdir\n\n";
  exit();
}
if($flagmand[1]==0){
  &print_help();
  print "please set -indir\n\n";
  exit();
}
if($flagmand[2]==0){
  &print_help();
  print "please set -pdb\n\n";
  exit();
}

if(!-s "$indir/$strname"){
  print "$strname doesn't exist, $!\n";
  exit();
}
else{
  if($strname=~/\./){
    $ID=substr($strname,0,rindex($strname,"."));
  }
  else{
    $ID=$strname;
  }
}

if($flagmand[3]==0){
  $libdir=$pkgdir."/PDBlib";
}
if($flagmand[4]==0){
  if($outdir=~/$ID/){
    $outdir=$indir."/decoys";
  }
  else{
    $outdir=$indir."/$ID";
  }
  if(!-d $outdir){
    mkdir $outdir or die "Cannot mkdir $outdir, $!\n";
  }
}
elsif(!($outdir=~/$ID/)){
  $outdir=$outdir."/$ID";
  if(!-d $outdir){
    mkdir $outdir or die "Cannot mkdir $outdir, $!\n";
  }
}

$user=getlogin() || $ENV{'USER'};


if($runstyle ne "parallel"){
  $runstyle="serial";
  $npara=1;
}
else{
  if ( !( `which qstat` ) || !( `which qsub` )){
    print "\nYou can not set \"-run =parallel\" because your system dose not support the commands: qsub and qstat.\n";
    print "To support qsub and qstat, you have to install the job management software PBS server.\n";
    exit();
  }
}

################ arguments printing ###################

if($argshow eq "true"){
  printf "========================================================\n";
  printf "The setting for running 3DRobot is:\n";
  printf "-pkgdir   = $pkgdir\n";
  printf "-indir    = $indir\n";
  printf "-pdb      = $strname\n";
  printf "-libdir   = $libdir\n";
  printf "-outdir   = $outdir\n";
  printf "-run      = $runstyle\n";
  printf "-npara    = $npara\n";
  printf "-nd       = $ndecoy\n";
  printf "-cut      = $rmsdcut\n";
  printf "-hour     = $hours\n";
  printf "-record   = $rcd\n";
  printf "-check    = $check\n";
  printf "========================================================\n";
}

$ttm1=time;
#################### basic setting ######################
$bindir=$pkgdir."/bin";
$commondir=$pkgdir."/common";
$refdir=$pkgdir."/refinement";
$recorddir=$indir."/record_$ID";
if(!-d $recorddir){
  mkdir $recorddir or die "Cannot mkdir $recorddir, $!\n";
}

$binwd=1.0; #RMSD bin width.
$nbin=int($rmsdcut/$binwd);
$nbdcy=int($ndecoy/$nbin); 
$remdcy=$ndecoy%$nbin;
$max_rerun=1;

######## generate basic files based on input structure #####
chdir $recorddir or die "Cannot chdir to $recorddir, $!\n";

@basefile=qw(seq.fasta rmsinp CA);
if($check eq "true" && (!-s $basefile[0] || !-s $basefile[1] || !-s $basefile[2]) || $check eq "false"){
  print "Step 1: make seq.fasta, rmsinp and CA\n";
  `cp $indir/$strname $recorddir/`;
  `$bindir/prepro.pl $strname $bindir`;
}

###################### Run TMalign ######################
$init="init.dat";

if($check eq "true" && (!-s "$init") || $check eq "false"){

  print "Step 2: Run TMalign ($init)\n";
  $TMali="TMalimod";
  $tag="TMali\_$ID";
  $mod=`cat $bindir/$TMali`;
  $mod=~s/\!TAG\!/$tag/mg;
  $mod=~s/\!USER\!/$user/mg;
  $mod=~s/\!BINDIR\!/$bindir/mg;
  $mod=~s/\!RECORDDIR\!/$recorddir/mg;
  $mod=~s/\!LIBDIR\!/$libdir/mg;
  $mod=~s/\!INIT\!/$init/mg;
  $mod=~s/\!STRNAME\!/$strname/mg;
  $mod=~s/\!NDECOY\!/$ndecoy/mg;
  $mod=~s/\!RMSDCUT\!/$rmsdcut/mg;
  open(JOB,">$recorddir/$TMali");
  print JOB "$mod\n";
  close(JOB);
  `chmod 755 $recorddir/$TMali`;
  system("$recorddir/$TMali");
#---- check if TMalign finished ----
  if(!-s "$recorddir/$init"){
    print "No output from TMalign, $!\n";
    exit();
  }
}

############## derive second structure info ##############
if($check eq "true" && (!-s "seq.stru") || $check eq "false"){

  print "Step 3: make seq.stru and ssdis.dat\n";
  `$bindir/secinfo.pl $libdir $bindir $recorddir $init $strname`;
  if(!-s "$recorddir/seq.stru" || !-s "$recorddir/ssdis.dat"){
    print "No output from secinfo.pl, $!\n";
    exit();
  }
}

################### run MC simulation ####################
open(INI,"$recorddir/$init");
if(<INI>=~/(\S+)\s+\S+\s+\S+/){
  $n_tmp=$1; #number of templates
}
close(INI);

$flag=0;
for($i=1;$i<=$n_tmp;$i++){
  if($check eq "true" && (!-s "$recorddir/trajs$i.bz2") || $check eq "false"){
    print "Cannot find trajs$i.bz2, re-run simulation.\n" if $check eq "true";
    $flag=1;
    last;
  }
}

if($flag==1){
  $tm1=time;
  print "Step 4: Run MC simulation\n";

  $shr=int($hours/3);

  $ncycle=100+int(200*sqrt($ndecoy/1000.0));
  $nrep=30+int(10*sqrt($ndecoy/1000.0));

  $hour=2;
  $hr=int(8*sqrt($ndecoy/1000.0)) if $ndecoy<=1000;
  $hour=$hr if $hr>$hour;
  open(RMSI,"rmsinp") or die "Cannot open rmsinp $!\n";
  <RMSI>=~/\d+\s+(\d+)/;
  $Lch=$1;
  close(RMSI);
  if($Lch>180 && $Lch<300){
    $hr=int(8*($Lch/300.0)**2);
    $hour=$hr if $hr>$hour;
  }elsif($Lch>=300){
    $hour=8;
  }
  $hour=$shr if $hour>$shr;

  $ncycle=int($ndecoy*2.0/$n_tmp) if $ndecoy*2.0/$n_tmp>$ncycle;
  $ncycle=500 if($ncycle>500);


  `/bin/rm -f $recorddir/trajinfo*.dat`;
  `/bin/rm -f $recorddir/trajs*.bz2`;
  `/bin/rm -f $recorddir/out*`;

  $npara=$n_tmp if $npara>$n_tmp;
  $nint=int($n_tmp/$npara);
  $nrem=$n_tmp%$npara;
  $tlast=0;

  $max_rerun=3 if $npara>1;
  %parajob=();
  for($i=1;$i<=$npara;$i++){
    $parajob{$i}=$i;
  }
  $njob=$npara;
  while($njob>0 && $max_rerun>0){
    for($i=1;$i<=$npara;$i++){
      next unless exists $parajob{$i};
      $tfirst=$tlast+1;
      $tlast=$tfirst+$nint-1;
      $tlast++ if $i<=$nrem;

      `rm $recorddir/out$i` if -s "$recorddir/out$i";

      $tag="Dcas$ID\_$i";
      $mod=`cat $bindir/casmod`;
      $mod=~s/\!I\!/$i/mg;
      $mod=~s/\!TAG\!/$tag/mg;
      $mod=~s/\!USER\!/$user/mg;
      $mod=~s/\!RECORDDIR\!/$recorddir/mg;
      $mod=~s/\!BINDIR\!/$bindir/mg;
      $mod=~s/\!COMMONDIR\!/$commondir/mg;
      $mod=~s/\!INIT\!/$init/mg;
      $mod=~s/\!NCYCLE\!/$ncycle/mg;
      $mod=~s/\!HOURS\!/$hours/mg;
      $mod=~s/\!NREP\!/$nrep/mg;
      $mod=~s/\!NBIN\!/$nbin/mg;
      $mod=~s/\!BINWD\!/$binwd/mg;
      $mod=~s/\!NBDCY\!/$nbdcy/mg;
      $mod=~s/\!TFIRST\!/$tfirst/mg;
      $mod=~s/\!TLAST\!/$tlast/mg;

      $jobname="Dcas$ID\_$i";
      open(JOB,">$recorddir/$jobname");
      print JOB "$mod\n";
      close(JOB);
      `chmod 755 $recorddir/$jobname`;

      ### submit job #2-#$npara
      next if($i==1);
      open(OUTPUT,">$recorddir/$jobname\.pbs");
      printf OUTPUT "#PBS -N Dcas$ID\_$i\n";
      printf OUTPUT "#PBS -l nodes=1:ppn=1,mem=1000mb,walltime=48:00:00\n";
      printf OUTPUT "#PBS -o $recorddir/$jobname".".out\n";
      printf OUTPUT "#PBS -e $recorddir/$jobname".".err\n";
      printf OUTPUT "$recorddir/$jobname\n";
      close(OUTPUT);
      
      $bsub="";
      while(length $bsub==0){
        $bsub=`qsub $recorddir/$jobname\.pbs`;
        chomp($bsub);
        sleep(10) if length $bsub==0;
      }
    }
    ### run job #1
    if(exists $parajob{1}){
      $jobname="Dcas$ID\_1";
      `$recorddir/$jobname`;
    }

    ### check if all MC simulations complete.
    if($npara>1){
      while(1){
        $counter=`qstat -u $user |grep Dcas$ID |wc -l`;
        $counter=~s/\s//g;
        last if $counter<=0;
        sleep(30);
      }
    }
    sleep(2);
    $njob=$npara;
    for($i=1;$i<=$npara;$i++){
      if(-s "$recorddir/out$i"){
        $njob--;
        delete $parajob{$i} if exists $parajob{$i};
      }
    }
    if($njob>0){
      $max_rerun--;
      print "$njob Simu-job has no output, Rerun!\n" if $max_rerun>0;
    }
  }

  $tm2=time;
  $tmcost=timecost($tm1,$tm2);
  print "Simulation cost: $tmcost\n";
}

####################### decoy selection #####################
if($check eq "true" && (!-s "$recorddir/$ID.tar.bz2") || $check eq "false"){

  print "Step 5: Select decoys from C-alpha trajectories\n";
  $selecmod="selecmod";
  $tag="selec\_$ID"; 
  $mod=`cat $bindir/$selecmod`;
  $mod=~s/\!TAG\!/$tag/mg;
  $mod=~s/\!USER\!/$user/mg;
  $mod=~s/\!BINDIR\!/$bindir/mg;
  $mod=~s/\!RECORDDIR\!/$recorddir/mg;
  $mod=~s/\!RMSDCUT\!/$rmsdcut/mg;
  $mod=~s/\!BINWD\!/$binwd/mg;
  $mod=~s/\!REFDIR\!/$refdir/mg;
  $mod=~s/\!NBDCY\!/$nbdcy/mg;
  $mod=~s/\!REMDCY\!/$remdcy/mg;
  $mod=~s/\!STRNAME\!/$strname/mg;
  $mod=~s/\!SS\!/$ID/mg;
  open(JOB,">$recorddir/$selecmod");
  print JOB "$mod\n";
  close(JOB);
  `chmod 755 $recorddir/$selecmod`;
  system("$recorddir/$selecmod");

  if(!-s "$recorddir/$ID.tar.bz2"){
    print "No output from selecmod, $!\n";
    exit();
  }
}

####################### decoy refinement #####################
if($check eq "true" && (!-s "$recorddir/refined.txt") || $check eq "false"){
  print "Step 6: Decoy refinement\n";
  $tm1=time;

  `/bin/rm -f $outdir/*`;
  `/bin/cp -f $recorddir/$strname $outdir`;

  $max_rerun=3 if $npara>1;
  %parajob=();
  for($i=1;$i<=$npara;$i++){
    $parajob{$i}=$i;
  }
  $njob=$npara;
  while($njob>0 && $max_rerun>0){
    for($i=1;$i<=$npara;$i++){
      next unless exists $parajob{$i};
      $tag="Dref$ID\_$i";
      $mod=`cat $bindir/refinemod`;
      $mod=~s/\!IPARA\!/$i/mg;
      $mod=~s/\!TAG\!/$tag/mg;
      $mod=~s/\!USER\!/$user/mg;
      $mod=~s/\!RECORDDIR\!/$recorddir/mg;
      $mod=~s/\!OUTDIR\!/$outdir/mg;
      $mod=~s/\!HOURS\!/$hours/mg;
      $mod=~s/\!BINDIR\!/$bindir/mg;
      $mod=~s/\!REFDIR\!/$refdir/mg;
      $mod=~s/\!NPARA\!/$npara/mg;
      $mod=~s/\!STRNAME\!/$strname/mg;
      $mod=~s/\!SS\!/$ID/mg;

      $jobname="Dref$ID\_$i";
      open(JOB,">$recorddir/$jobname");
      print JOB "$mod\n";
      close(JOB);
      `chmod 755 $recorddir/$jobname`;

      ### submit job #2-#$npara
      next if($i==1);
      open(OUTPUT,">$recorddir/$jobname\.pbs");
      printf OUTPUT "#PBS -N Dref$ID\_$i\n";
      printf OUTPUT "#PBS -l nodes=1:ppn=1,mem=1000mb,walltime=99:00:00\n";
      printf OUTPUT "#PBS -o $recorddir/$jobname".".out\n";
      printf OUTPUT "#PBS -e $recorddir/$jobname".".err\n";
      printf OUTPUT "$recorddir/$jobname\n";
      close(OUTPUT);
      
      $bsub="";
      while(length $bsub==0){
        $bsub=`qsub $recorddir/$jobname\.pbs`;
        chomp($bsub);
        sleep(10) if length $bsub==0;
      }
    }
    ### run job #1
    if(exists $parajob{1}){
      $jobname="Dref$ID\_1";
      `$recorddir/$jobname`;
    }

    ### check if all refinement jobs complete.
    if($npara>1){
      while(1){
        $counter=`qstat -u $user |grep Dref$ID |wc -l`;
        $counter=~s/\s//g;
        last if $counter<=0;
        sleep(30);
      }
    }
    sleep(2);
    $njob=$npara;
    for($i=1;$i<=$npara;$i++){
      if(-s "$recorddir/REF$i"){
        $njob--;
        delete $parajob{$i} if exists $parajob{$i};
      }
    }
    if($njob>0){
      $max_rerun--;
      print "$njob ref-job has no output, Rerun!\n" if $max_rerun>0;
    }
  }

  system("$bindir/collec.pl $ID $recorddir $outdir $user");
  $tm2=time;
  $tmcost=timecost($tm1,$tm2);
  print "refinement cost: $tmcost\n";
}

######### evaluate decoys and generate rst.dat ##########
$rst="rst.dat";
if($check eq "true" && (!-s "$outdir/$rst") || $check eq "false"){

  print "Step 7: Decoy evaluation\n";
  $evamod="evamod";
  $tag="Deva\_$ID"; 
  $mod=`cat $bindir/$evamod`;
  $mod=~s/\!TAG\!/$tag/mg;
  $mod=~s/\!USER\!/$user/mg;
  $mod=~s/\!BINDIR\!/$bindir/mg;
  $mod=~s/\!RECORDDIR\!/$recorddir/mg;
  $mod=~s/\!OUTDIR\!/$outdir/mg;
  $mod=~s/\!RMSDCUT\!/$rmsdcut/mg;
  $mod=~s/\!BINWD\!/$binwd/mg;
  $mod=~s/\!STRNAME\!/$strname/mg;
  $mod=~s/\!SS\!/$ID/mg;
  open(JOB,">$recorddir/$evamod");
  print JOB "$mod\n";
  close(JOB);
  `chmod 755 $recorddir/$evamod`;
  system("$recorddir/$evamod");
}

################# endding procedure ######################
if($rcd ne "true"){
  chdir $bindir;
  `rm -fr $recorddir`;
}

$time=`date`;
printf "3DRobot ending time: $time";

$ttm2=time;
$tmcost=timecost($ttm1,$ttm2);
print "Total time: $tmcost\n";
`sync`;
exit();

sub timecost{
  $hor=int(($_[1]-$_[0])/3600);
  $min=int((($_[1]-$_[0])%3600)/60);
  $sec=($_[1]-$_[0])%60;
  $tmcst="$hor h $min m $sec s";
  return $tmcst;
}

sub print_help
{
  printf 
"3DRobot Usage:
=====================
Mandatory arguments:
=====================
./run3DRobot.pl -pkgdir package_dir -indir input_dir -pdb input_structure
=====================
Optional arguments:
=====================
  -libdir    the path of structure template library, default is \"package_dir/PDBlib\"
  -outdir    the path where output decoys save
  -run       [serial, parallel], \"serial\"(default), run 3DRobot sequentially
             \"parallel\", run 3DRobot simulations/refinements in parallel, distributed on cluster nodes
  -npara     number of allowed parallel jobs (default=10 when -run=parallel)
  -nd        number of output decoys, default=100, recommended range is in [10,1000]
  -cut       RMSD cutoff for output decoys, default=12, recommended ragne [5,20]
  -hour      maximum runing time for simulations&refinements (default=48, and minimum=6)
  -check     [true or false], \"true\", check the existing files and start from the correct position
             \"false\"(default), run all steps without check
  -record    [true or false], \"true\", keep recod&intermediate files, \"false\"(default), do not keep them
  -args      [true or false], \"true\"(default), show the setting informations, \"false\", do not show them\n";
}