#!/usr/bin/perl -w
#############################
# All-atom decoy generation #
#############################

$strname=$ARGV[0];
$rmsdcut=$ARGV[1];
$binwd=$ARGV[2];
$nbdcy=$ARGV[3];
$remdcy=$ARGV[4];
$ss=$ARGV[5];
$dlst="list.txt";
$ca="CA";
$natdcy=1; # when=1, add a decoy which is built from C-alpha of input structure.

@dname=@rmsds=();

$nbin=int($rmsdcut/$binwd);

if($ss eq $ca){
  rename($ca,"CAlpha");
  $ca="CAlpha";
}
mkdir("$ss") if !-s "$ss";
open(LST,"> $dlst") or die "Can't make:$dlst $!\n";
printf LST "NAME                  RMSD\n";
printf LST "%-18s    0.00\n",$strname;

######## read sequence length from rmsinp #########
open(RMSI,"rmsinp") or die "Cannot open rmsinp $!\n";
<RMSI>=~/\d+\s+(\d+)/;
$Lch=$1;
close(RMSI);

######## read sequence infomation #########
$lchk=0;
%seq=();
%alterp=();
open(PDB,"$strname") or die "Cannot open $strname $!\n";
while(<PDB>){
  if($_=~/^ATOM/){
    $check=substr($_,22,4).substr($_,12,4);
    if(!exists $alterp{$check}){
      if(substr($_,13,2) eq 'CA'||substr($_,12,2) eq 'CA'||substr($_,14,2) eq 'CA'){
        $amino=substr($_,17,3);
        $nseq=substr($_,22,4);
        $nseq=~s/\s+//;
        $seq{$nseq}=$amino;
        $lchk++;
      }
      $alterp{$check}="Done";
    }
  }elsif($_=~/^TER/){
    last;
  }
}
close(PDB);

if($lchk!=$Lch){
  printf "Protein length inconsistence in rmsinp($Lch) and $strname($lchk).\n";
  exit;
}

$natsecp=&secpct("$strname");

####### select trajectories according to trajinfo*.dat #########
opendir(DD, "./") or die "Can't open:./ $!\n" ;
@files= readdir(DD);
closedir(DD);

$m=$n=0;
%trinfo=();
%traj=();
foreach $fl (@files){
  if($fl=~/trajinfo(\d+).dat/){
    $m++;
    $trinfo{$1}=$fl;
  }elsif($fl=~/trajs(\d+).bz2/){
    $n++;
    $traj{$1}=substr($fl,0,index($fl,"."));
  }
}
@tron= keys %trinfo;
if($m!=$n){
  print "Waining: #of trajinfo and trajs files are inconsistent.\n";
  foreach(@tron){
    delete $trinfo{$_} if !exists $traj{$_};
  }
  @tron= keys %trinfo;
}

#>>> read #of traj in each bin of each file.
%ndcy=();
$ntmp=@tron;
foreach $tr (@tron){
  open(TRIN,"$trinfo{$tr}") or die "Can't open:$trinfo{$tr} $!\n";
  while($line=<TRIN>){
    if($line=~/^n_dcy /){
      for($i=0;$i<$nbin;$i++){
        $ndcy{$tr}{$i}=substr($line,6+$i*7,7);
      }
      last;
    }
  }
  close(TRIN);
}

#>>> check every bin of every file and select traj.
%binchk=(); #record the bins which have decoys left, also the number of decoys.
for($i=0;$i<$nbin;$i++){
  $binchk{$i}=0;
}
$totlft=0;
$nrems=$remdcy;
%nslc=();
for($i=0;$i<$nbin;$i++){
  $nbtot=0;
  foreach $tr (@tron){
    $nbtot=$nbtot+$ndcy{$tr}{$i};
  }
  if($nbtot>=$nbdcy){
    $nbslc=0;
    foreach $tr (@tron){
      if($ndcy{$tr}{$i}>0){
        $nslc{$tr}{$i}=int($nbdcy*$ndcy{$tr}{$i}/$nbtot);
        $nslc{$tr}{$i}=1 if $nslc{$tr}{$i}<1 && $nbslc<$nbdcy;
        $nslc{$tr}{$i}=$nbdcy-$nbslc if ($nbslc+$nslc{$tr}{$i})>$nbdcy;
      }else{
        $nslc{$tr}{$i}=0;
      }
      $nbslc=$nbslc+$nslc{$tr}{$i};
    }
    $nbrem=$nbdcy-$nbslc;

    @tmp= sort {$ndcy{$b}{$i} <=> $ndcy{$a}{$i}} @tron;
    $j=0;
    while($nbrem>0){
      $lft=$ndcy{$tmp[$j]}{$i}-$nslc{$tmp[$j]}{$i};
      if($lft>0){
        $nslc{$tmp[$j]}{$i}++;
        $nbrem--;
      }
      $j++;
      $j=0 if $j>=$ntmp;
    }
    $binchk{$i}=$nbtot-$nbdcy;
    $totlft=$totlft+$binchk{$i};
  }else{
    foreach $tr (@tron){
      $nslc{$tr}{$i}=$ndcy{$tr}{$i};
    }
    $nrems=$nrems+($nbdcy-$nbtot);
    delete $binchk{$i};
  }
}
$nrems=$totlft if $totlft<$nrems;


while($nrems>0){
  @bchk=keys %binchk;
  $nblft=@bchk;
  if($nrems>=$nblft){
    $av=int($nrems/$nblft);
    foreach $i (@bchk){
      if($binchk{$i}<=$av){
        foreach $tr (@tron){
          $nslc{$tr}{$i}=$ndcy{$tr}{$i};
        }
        $nrems=$nrems-$binchk{$i};
        delete $binchk{$i};
      }else{
        $nbslc=0;
        foreach $tr (@tron){
          $trblft=$ndcy{$tr}{$i}-$nslc{$tr}{$i};
          if($trblft>0){
            $nslc{$tr}{$i}=$nslc{$tr}{$i}+int($av*$trblft/$binchk{$i});
            $nbslc=$nbslc+int($av*$trblft/$binchk{$i});
          }
        }
        $nbrem=$av-$nbslc;

        @tmp= sort {$ndcy{$b}{$i} <=> $ndcy{$a}{$i}} @tron;
        $j=0;
        while($nbrem>0){
          $lft=$ndcy{$tmp[$j]}{$i}-$nslc{$tmp[$j]}{$i};
          if($lft>0){
            $nslc{$tmp[$j]}{$i}++;
            $nbrem--;
          }
          $j++;
          $j=0 if $j>=$ntmp;
        }
        $binchk{$i}=$binchk{$i}-$av;
        $nrems=$nrems-$av;
      }
    }
  }else{
    foreach $i (@bchk){
      foreach $tr (@tron){
        if($nslc{$tr}{$i}<$ndcy{$tr}{$i}){
          $nslc{$tr}{$i}++;
          $nrems--;
          last;
        }
      }
      last if $nrems<=0;
    }
  }
}

$nmax=0;
for($i=0;$i<$nbin;$i++){
  $binslc[$i]=0;
  foreach $tr (@tron){
    $binslc[$i]+=$nslc{$tr}{$i};
    if($nslc{$tr}{$i}>$nmax && $i>0){
      $nmax=$nslc{$tr}{$i};
      $mtr=$tr;
      $mi=$i;
    }
  }
}

######### add decoy built from the backbone of input structure  ###########
if($natdcy==1){
  $nslc{$mtr}{$mi}--;
  $binslc[$mi]--;
  if(-s "$ca"){
    `./pulchra304 -pqs $ca`;

    $result=`./RMSD $ca.rebuilt.pdb $ca`;
    $RMSD=$1 if $result=~/RMSD of the common residues=\s+(\S+)/;
    #`cp $ca.rebuilt.pdb $ss/decoy0_1.pdb`;
    #`./Scwrl4 -i $ca.rebuilt.pdb -o $ss/decoy0_1.pdb`;
    $rand=int(rand()*1000000);
    `./emrefinement ./ ./ $ca.rebuilt.pdb $ca.rebuilt.pdb 100 $rand 0`;
    `cp em$ca.rebuilt.pdb $ss/decoy0_1.pdb`;
    printf LST "%-18s%8.2f\n","decoy0_1.pdb",$RMSD;
  }else{
    print "Warning: cannot find $ca file\n";
  }
}
################ trajectories to decoys ##################
for($i=0;$i<$nbin;$i++){
  $evnss[$i]=0;
}
$evnss[$nbin]=0;

`bzip2 -d trajs*.bz2`;

$ndsec=0;
$avesecp=0.0;
@dsecs=();
@enegy=();
@trmsd=();
for($i=0;$i<=$nbin;$i++){
  $ntr[$i]=0;
}
@tron= sort {$a <=> $b} @tron;
foreach $tr (@tron){
  open(TRA, "$traj{$tr}") or die "Can't open $traj{$tr} $!\n" ;
  #print "$traj{$tr} is calculating...\n";
  while($line=<TRA>){
    if($line=~/\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/){
      $n=$1;
      $energy=$2;
      $rmsd=$3;
      $bin=$4;
      $trn=$5;

      $dcynam="decoy$tr\_$trn";
      $dsecs[$bin][$ntr[$bin]]=0.0;
      $enegy[$bin][$ntr[$bin]]=$energy;
      $trmsd[$bin][$ntr[$bin]]=$rmsd;
      $ntrtmp[$bin][$ntr[$bin]]=$tr;
      $ntrnam[$bin][$ntr[$bin]]=$dcynam;

      $m=0;
      $RMSD=0;
      @x=@y=@z=();
      open(DECOY, "> $dcynam") or die "Can't make $dcynam $!\n" ;
      next;
    }
    if($line=~/\s+(\S+)\s+(\S+)\s+(\S+)/){
      $x[$m]=$1;
      $y[$m]=$2;
      $z[$m]=$3;
      printf DECOY "ATOM  %5d  CA  %3s  %4d    %8.3f%8.3f%8.3f\n",$m+1,$seq{$m+1},$m+1,$x[$m],$y[$m],$z[$m];
      $m++;
      if($m>=$n){
        print DECOY "TER  ";
        close(DECOY);
        if($m!=$Lch){
          print "length of traj $trn in $traj{$tr} is inconsistent with input structure. quit!\n";
          exit();
        }
        `./pulchra304 -sc $dcynam`;
        if(-s "$dcynam.rebuilt.pdb"){
          $dsecs[$bin][$ntr[$bin]]=&secpct("$dcynam.rebuilt.pdb");
          $avesecp+=$dsecs[$bin][$ntr[$bin]];
          $ntr[$bin]++;
          $ndsec++;
        }else{
          print "Warning: no output $dcynam.rebuilt.pdb\n";
          exit();
        }
      }
    }
  }
  close(TRA);
}
$avesecp/=$ndsec;


################ decoy selection ################
if($natsecp>$avesecp*0.95){
  if($natsecp<40 && $avesecp>$natsecp*0.6){
    $slctmthd=0;
  }else{
    $slctmthd=1; #select decoys according to SecPct regardless temp based.
  }
}else{
  $slctmthd=0;
}

%tmpused=();
for($bin=0;$bin<$nbin;$bin++){
  next if $ntr[$bin]<=0;
  @rank=();
  for($itr=0;$itr<$ntr[$bin];$itr++){
    $rank[$itr]=$itr;
  }
  @rank= sort {$dsecs[$bin][$b] <=> $dsecs[$bin][$a]} @rank;
  if($slctmthd==0){
    for($itr=0;$itr<$ntr[$bin];$itr++){
      if($nslc{$ntrtmp[$bin][$rank[$itr]]}{$bin}>0){
        &dslct($rank[$itr],$bin);
        $nslc{$ntrtmp[$bin][$rank[$itr]]}{$bin}--;
        $binslc[$bin]--;
      }
    }
  }else{
    %dslcted=();
    for($itr=0;$itr<$ntr[$bin];$itr++){
      if(!exists $tmpused{$ntrtmp[$bin][$rank[$itr]]} && $binslc[$bin]>0){
        $tmpused{$ntrtmp[$bin][$rank[$itr]]}="used";
        &dslct($rank[$itr],$bin);
        $dslcted{$rank[$itr]}="slcted";
        $nslc{$ntrtmp[$bin][$rank[$itr]]}{$bin}--;
        $binslc[$bin]--;
      }
    }
    for($itr=0;$itr<$ntr[$bin];$itr++){
      if(!exists $dslcted{$rank[$itr]} && $binslc[$bin]>0){
        &dslct($rank[$itr],$bin);
        $dslcted{$rank[$itr]}="slcted";
        $nslc{$ntrtmp[$bin][$rank[$itr]]}{$bin}--;
        $binslc[$bin]--;
      }
    }
  }
}

################ generate decoy list ################
$nd=@rmsds;
for($i=0;$i<$nd;$i++){
  $rk[$i]=$i;
}
for($i=0;$i<$nd-1;$i++){
  for($j=$nd-1;$j>$i;$j--){
    $jb=$1 if $dname[$rk[$j]]=~/decoy(\d+)/;
    $js=$1 if $dname[$rk[$j-1]]=~/decoy(\d+)/;
    if($jb<$js){
      $tmp=$rk[$j];
      $rk[$j]=$rk[$j-1];
      $rk[$j-1]=$tmp;
    }
  }
}

%ntemp=();
for($i=0;$i<$nd;$i++){
  printf LST "%-18s%8.2f\n",$dname[$rk[$i]],$rmsds[$rk[$i]];
  if($dname[$rk[$i]]=~/decoy(\d+)/){
    if(exists $ntemp{$1}){
      $ntemp{$1}++;
    }else{
      $ntemp{$1}=1;
    }
  }
}
close(LST);
`cp $dlst $ss/`;
`cp $strname $ss/`;
`tar jcvf $ss.tar.bz2 $ss/`;
`rm $ss.tar.bz2` if -s "$ss.tar.bz2" <100;

####################### Program end #######################


sub dslct{
  $bd=$_[0];
  $rb=$_[1];
  $dcynam=$ntrnam[$rb][$bd];
  `./pulchra304 -pqs $dcynam`;
  if(-s "$dcynam.rebuilt.pdb"){
    $result=`./RMSD $dcynam.rebuilt.pdb $ca`;
    $RMSD=$1 if $result=~/RMSD of the common residues=\s+(\S+)/;
    #`cp $dcynam.rebuilt.pdb $ss/$dcynam.pdb`;
    #`./Scwrl4 -i $dcynam.rebuilt.pdb -o $ss/$dcynam.pdb`;
    $rand=int(rand()*1000000);
    `./emrefinement ./ ./ $dcynam.rebuilt.pdb $dcynam.rebuilt.pdb 100  $rand 0`;
    `cp em$dcynam.rebuilt.pdb $ss/$dcynam.pdb`;
    push(@dname,"$dcynam.pdb");
    push(@rmsds,$RMSD);
  }else{
    print "Warning: no output $dcynam.rebuilt.pdb\n";
  }
}

sub secpct{
  $nsec=0;
  $totsec=0;
  @results=`./stride $_[0]`;
  foreach $result (@results){
    if($result=~/ASG\s+\S+ \S+\s+(\d+)\s+\d+\s+(\S+)/){
      $i=$1;
      $j=$2;
      if($j eq "H"||$j eq "G" ||$j eq "I"){
        $sc=2;
      }
      elsif($j eq "B" ||$j eq "b" ||$j eq "E"){
        $sc=4;
      }else{
        $sc=1;
      }
      if($sc>1){
        $nsec++;
      }
      $totsec++;
    }
  }
  $spct=$nsec*100.0/$totsec;
  return $spct;
}